﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________  													//
//																						//
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt16 prgLdrMode __attribute__ ((section (".noinit")));

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define FreezeWhileRegBitClr(reg, bit, timout) FreezeWhileRegBitClr_D(&(reg), bit, timout)
__attribute__((always_inline)) static inline uint8_t
FreezeWhileRegBitClr_D(volatile uint8_t* sfr, uint8_t bit, uint8_t timout)
{
    uint8_t            mask;
    volatile uint16_t  bailOut;

    mask    = 1 << bit;
    bailOut = (F_CPU/(1+CKDIV8*7)*timout)/(8*1000); // 8 = # Zyklen pro Iteration
    if (bailOut == 0) bailOut = 1;
    __asm__ __volatile__ ("L1_%=:  ld    __tmp_reg__, x   \n\t"
                          "        and   __tmp_reg__, %2  \n\t"
                          "        brne  L2_%=            \n\t"
                          "        sbiw  %0,1             \n\t"
                          "        brne  L1_%=            \n\t"
                          "L2_%=:                             "
                          : "+w" (bailOut)
                          : "x" (sfr),
                            "r" (mask)
                          );
    return bailOut > 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Vorlauf  /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                           _                                                          //
//                          (_)                                                         //
//                           |                                                          //
//                           |                                                          //
//                          \|/                                                         //
//                    /¯¯¯¯¯¯¯¯¯¯¯¯¯¯\  ja                                              //
//                   | Power-On-Reset |------------------+                              //
//                    \______________/                   |                              //
//                           | nein                      |                              //
//                           |                           |                              //
//                          \|/                         \|/                             //
//                    /¯¯¯¯¯¯¯¯¯¯¯¯¯¯\ nein       /¯¯¯¯¯¯¯¯¯¯¯¯¯¯\ ja                   //
//                   | External-Reset |--------->| Watchdog-Reset |------+              //
//                    \______________/            \______________/       |              //
//                           | ja                        | nein          |              //
//                           |                           |               |              //
//                          \|/                         \|/              |              //
//                +----------------------+   +-----------------------+   |              //
//                | mode = PGMLDR_MARKER |   | mode = !PGMLDR_MARKER |   |              //
//                +----------------------+   +-----------------------+   |              //
//                           |                           |               |              //
//                           |<--------------------------+---------------+              //
//                           |                                                          //
//                          \|/                                                         //
//                /¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯\  nein                                         //
//               | mode != PGMLDR_MARKER |-----------------+                            //
//                \_____________________/                  |                            //
//                           | ja                          |                            //
//                           |                             |                            //
//                          \|/                            |                            //
//                  /¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯\  nein              |                            //
//                 | Read(0) != 0xFFFF |------------------>|                            //
//                  \_________________/                    |                            //
//                           |  ja                         |                            //
//                           |                             |                            //
//                   ________|________            _________|__________                  //
//                   Enter Application            Enter Program Loader                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked)) __attribute__((section(".init3"))) void
PreStart(void)
{
    UInt08 mcusr = MCUSR;
    MCUSR = 0;
    UInt16 mode  = prgLdrMode;
    wdt_enable(WDTO_2S);
    if (mcusr & BV(PORF))
    {   mode = !CFG_PGMLDR_MARKER;
    }
    else
    {   if (mcusr & BV(EXTRF))
        {   mode = CFG_PGMLDR_MARKER;
        }
        else
        {   if (!(mcusr & BV(WDRF)))
            {   mode = !CFG_PGMLDR_MARKER;
            }
        }
    }
    if (mode != CFG_PGMLDR_MARKER)
    {   if (pgm_read_word_near(0) != 0xFFFF) GOTO(0);  // ... Enter Application
    }
    // ... Enter Program Loader
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(__vector_default) __attribute__((naked));
ISR(__vector_default)
{
    // Programmlader Einstiegspunkt
    DisableInterrupts();
    BOARD_Reset();
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Setup(void)
{
	// Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(B, _L_,    // PB.0: ---    : ---
                _L_,    // PB.1: BUZZ   : Piezosignalgeber, Frequenz über Timer 1 PWM
                _L_,    // PB.2: ---    : ---
                _L_,    // PB.3: MOSI   : ISP
                _L_,    // PB.4: MISO   : ISP
                _L_,    // PB.5: SCK    : ISP
                _T_,    // PB.6: XTAL1  : Quarz
                _T_);   // PB.7: XTAL2  : Quarz
    ConfPort(C, _L_,    // PC.0: ---    : ---
                _L_,    // PC.1: Sw_Set : Schaltausgang 24V, L = geöffnet
                _T_,    // PC.2: Ct_In  : Kontakt, H = geöffnet
                _L_,    // PC.3: ---    : ---
                _L_,    // PC.4: ---    : ---
                _L_,    // PC.5: ---    : ---
                _T_,    // PC.6: RST    : Reset
                _T_);   //
    ConfPort(D, _T_,    // PD.0: RxD    : Bus
                _T_,    // PD.1: TxD    : Bus
                _L_,    // PD.2: ---    : ---
                _H_,    // PD.3: SYS_LED: System-LED, H = Ein
                _L_,    // PD.4: ---    : ---
                _L_,    // PD.5: ---    : ---
                _L_,    // PD.6: ExtLED : Externe LED, H = Ein
                _L_);   // PD.7: ---    : ---
    ConfPort(E, _L_,    // PE.0: ---    : ---
                _L_,    // PE.1: ---    : ---
                _L_,    // PE.2: ---    : ---
                _L_,    // PE.3: ---    : ---
                _L_,    //
                _L_,    //
                _L_,    //
                _L_);   //
    // 16-Bit Timer 1 ohne IRQ zählen lassen (F_CPU÷1024)
    TCCR1B = BV(CS12) | BV(CS10);
    // UART: 19200 Baud, 8N1, async
    UCSR0B = BV(RXEN0) | BV(TXEN0);
	UBRR0  = F_CPU/16L/19200-1;
}

//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) void
BOARD_Reset(void)
{
    prgLdrMode = CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) void
BOARD_EnterAppl(void)
{
    prgLdrMode = !CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_TransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_WaitForByte(UInt08* byte, UInt08 timeout)
{
    if (!FreezeWhileRegBitClr(UCSR0A, RXC0, timeout)) return false;
    *byte = UDR0;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////

